## Henderson.py
## calculate a Henderson moving average
## adpated from this:
## https://markthegraph.blogspot.com/2014/06/henderson-moving-average.html
## see also
## https://www.abs.gov.au/websitedbs/d3310114.nsf/4a256353001af3ed4b2562bb00121564/5fc845406def2c3dca256ce100188f8e!OpenDocument

import numpy as np

def hmaSymmetricWeights(n):
    """ derive an n-term array of symmetric 'Henderson Moving Average' weights
        formula from ABS (2003), 'A Guide to Interpreting Time Series', page 41.
        returns a numpy array of symmetric Henderson weights indexed from 0 to n-1"""

    # calculate the constant denominator and terms
    m = int((n-1)//2) # the mid point - n must be odd
    m1 = (m+1)*(m+1)
    m2 = (m+2)*(m+2)
    d = float(8*(m+2)*(m2-1)*(4*m2-1)*(4*m2-9)*(4*m2-25))
    m3 = (m+3)*(m+3)

    # calculate the weights
    w = np.repeat(np.nan, n) # Actually indexed from 0 to n-1
    for j in range(m+1):
        j2 = j*j
        v = (315*(m1-j2)*(m2-j2)*(m3-j2)*(3*m2-11*j2-16))/d
        w[(m+j)] = v
        if j > 0:
            w[(m-j)] = v

    w.flags.writeable = False # let's make it quasi-immutable
    return (w)


def hmaAsymmetricWeights(m, w):
    """calculate the asymmetric end-weights

        w --> an array of symmetrical henderson weights (from above function)
        m --> the number of asymmetric weights sought; where m < len(w);

        returns a numpy array of asymmetrical weights, indexed from 0 to m-1;

        formula from Mike Doherty (2001), 'The Surrogate Henderson Filters in X-11',
        Aust, NZ J of Stat. 43(4), 2001, pp901-999; see formula (1) on page 903"""

    n = len(w) # the number of weights

    # - some quick sanity checks
    if m >= n:
        raise ValueError('The m argument must be less than n')
    if m <= int((n-1)//2):
        raise ValueError('The m argument must be greater than (n-1)/2')

    # --- let's build up Doherty's formula (1) from the top of page 903

    # - the second chunk of the formula
    sumResidual = w[range(m, n)].sum() / float(m)

    # - the last chunk of the formula
    sumEnd = 0.0
    for i in range(m+1, n+1):
        sumEnd += (float(i)-((m+1.0)/2.0)) * w[i-1] # w indexed from 0 to n-1

    # - the beta squared / sigma squared - formula at the bottom of page 904
    ic = 1.0
    if n >= 13 and n < 15:
        ic = 3.5
    elif n >= 15:
        ic = 4.5
    b2s2 = (4.0/np.pi)/(ic*ic)

    # - the gnarly bit in the middle of the formula
    denominator = 1.0 + ((m*(m-1.0)*(m+1.0) / 12.0 ) * b2s2)
    u = np.repeat(np.nan, m) # return series - created empty
    for r in range(m): # r ranges 0 to m-1; but the formulae assumes 1 to m
        numerator = ((r+1.0) - (m+1.0)/2.0) * b2s2
        # - finally putting it all together
        u[r] = w[r] + sumResidual + ( numerator / denominator ) * sumEnd

    u.flags.writeable = False # let's make it quasi-immutable
    return (u)


def henderson(series, n):
    """ Calculate an n-term Henderson Moving Average for the Series s
        Note: we blithely assume s is ordered, contiguous and without missing data"""

    # - some simple sanity checks
    if not isinstance(series, list):
        raise TypeError('The series argument should be list')
    if not isinstance(n, int):
        raise TypeError('The n argument must be an integer')
    if n < 5:
        raise ValueError('The n argument must be >= 5')
    if n % 2 == 0:
        raise ValueError('The n argument must be odd')
    if len(series) < n:
        raise ValueError('The series argument should be a Series longer than n')

    # - calculate the symmetric weights
    w = hmaSymmetricWeights(n)

    # preliminaries
    r = [np.nan] * len(series) # the empty return vehicle
    m = int((n-1)//2)
    l = len(series)

    # - and now move over the length of the series ...
    for i in range(len(series)) :
        if i < m:
            # --- head section of series
            u = hmaAsymmetricWeights(m+i+1, w)[::-1] # reverse - asymmetric to the left
            r[i] = (series[0:(i+m+1)] * u).sum()
        elif i + m >= l:
            # --- tail section of series
            u = hmaAsymmetricWeights(m+l-i, w)
            r[i] = (series[(i-m):l] * u).sum()
        else:
            # --- middle section of series
            r[i] = (series[(i-m):(i+m+1)] * w).sum()

    return (r)


### - test code
#--------------
# Check against Table 1 in B Quenneville and B Lefrancois (2001),
# "Implicit Forecasts in Musgrave Asymmetric Averages",
# Proceedings of the Annual Meeting of the American Statistical Association,
# August 5-9, 2001.
#--------------
#w = hmaSymmetricWeights(9)
#print(w)
#print(w.sum()) # should be one
#u = hmaAsymmetricWeights(7, w)
#print(u)
#print(u.sum()) # should be one
#--------------
#print (Henderson([1,2,3,4,5,6,7,1,2,3,4,5,6,7,1,2,3,4,5,6,7,1,2,3,4,5,6,7,1,2,3,4,5,6,7,1,2,3,4,5,6,7,1,2,3,4,5,6,7,1,2,3,4,5,6,7,], 13))


import math

num_types = ["<class 'int'>", "<class 'float'>"]

# replaces nan items from a series with a given value
def fillna(series, filler=0):
    result = []
    for i in series:
        if (str(type(i)) in num_types) and not math.isnan(i):
            result.append(i)
        else:
            result.append(filler)
    return result

# average of the series
def mean(series):
    s = series.copy()
    s = fillna(s, None)
    while None in s:
        s.remove(None)
    if (len(s)) == 0:
        return None
    return sum(s) / float(len(s))

# standard deviateion of the series
def standard_deviation(series):
    return variance(series) ** (1.0/2.0)

# statistal variance of the series
def variance(series):
    s = series.copy()
    while None in s:
        s.remove(None)
    while math.nan in s:
        s.remove(math.nan)
    if (len(s)) == 0:
        return None
    series_mean = mean(s)
    return sum([(x-series_mean)**2.0 for x in s]) / len(s)

# executes a rule against each item in a series and returns matches, rule should be a lambe
# matches(data, lambda x: x > 2)
# list comprehension is slower compared to loops for shorter lists but faster for longer lists
def matches(series, rule):
    return [i for i, item in enumerate(series) if rule(item)]

# applied a function to a series of values, formula should be a lambda
# f_x(data, lambda x: 3x + 2)
def f_x(series, formula):
    return [formula(item) for item in series]

def series_diff(series_a, series_b, adjustment=None):
    if len(series_a) != len (series_b):
        raise Exception('series_diff: two series must be the same length')
    if adjustment == None:
        adjustment = mean(series_a)
    series = []
    for i in range(len(series_a)):
        series.append(series_a[i] - series_b[i] + adjustment)
    return series



from .general import *

class segment: 
    
    def __init__(self, movingrange, segment_start, segment_end, samples = 8):
        self.mR_series = []
        self.samples = samples
        self.start = segment_start
        self.end = segment_end
        self.mr = movingrange
        self.value_series = movingrange.value_series[segment_start:segment_end]
        self.mR_series = [abs(self.value_series[i+1]-self.value_series[i]) for i in range(len(self.value_series)-1)]

    def moving_range_mean(self):
        return mean(self.mR_series[0:self.samples])

    def moving_range_sigma(self):
        # mean x 3.267 = mean + 3 x sigma => sigma = (mean x 2.267) / 3
        mr_mean = self.moving_range_mean()
        sigma = (mr_mean * 2.267) / 3
        return sigma

    def moving_range_standard_deviation(self, number = 1):
        mr_mean = self.moving_range_mean()
        sigma = self.moving_range_sigma()
        std_dev = mr_mean + (sigma * number)
        if std_dev < 0:
            std_dev = 0
        return std_dev

    # puts mRs into a bin corresponding to their displacement from the mean in standard deviations
    def moving_range_bins(self):
        bins = []
        mr_mean = self.moving_range_mean()
        sigma = self.moving_range_sigma()
        for value in range(len(self.mR_series)):
            if sigma == 0:
                bin = 0
            else:
                bin = binner((self.value_series[value] - i_mean) / sigma)
            bins.append(bin)
        return bins

    # returns an array with items of 1 (increase), 0 (stable) or -1 (reduction)
    def moving_range_direction(self):
        directions = []
        for i in range(len(self.mR_series) - 1):
            delta = self.mR_series[i + 1] - self.mR_series[i]
            if delta == 0:
                direction = 0
            else:
                direction = delta / abs(delta)
            directions.append(direction)
        return directions

    def moving_range_sigma_line(self, number = 1):
        value = self.moving_range_standard_deviation(number)
        return [value] * len(self.value_series)

    def individuals_mean(self):
        return mean(self.value_series[0:self.samples])

    def individuals_sigma(self):
        mr_mean = self.moving_range_mean()
        sigma = (2.66 * mr_mean) / 3
        return sigma

    def individuals_standard_deviation(self, number = 1):
        i_mean = self.individuals_mean()
        sigma = self.individuals_sigma()
        std_dev = i_mean + (sigma * number)
        return std_dev

    # puts observations into a bin corresponding to their displacement from the mean in standard deviations
    def individuals_bins(self):
        bins = []
        i_mean = self.individuals_mean()
        sigma = self.individuals_sigma()
        for value in range(len(self.value_series)):
            if sigma == 0:
                bin = 0
            else:
                bin = binner((self.value_series[value] - i_mean) / sigma)
            bins.append(bin)
        return bins

    # returns an array with items of 1 (increase), 0 (stable) or -1 (reduction)
    def individuals_direction(self):
        directions = []
        for i in range(len(self.value_series) - 1):
            delta = self.value_series[i + 1] - self.value_series[i]
            if delta == 0:
                direction = 0
            else:
                direction = delta / abs(delta)
            directions.append(direction)
        return directions

    def describe(self):
        i_mean = self.individuals_mean()
        sigma = self.individuals_sigma()
        bins = self.individuals_bins()

        description = 'Segment' + '\n'
        description = description + '===========================' + '\n'
        description = description + 'Number of Samples = ' + str(len(self.value_series)) + '\n'
        description = description + 'Start Position = ' + str(self.start) + '\n'
        description = description + 'End Position = ' + str(self.end) + '\n'
        description = description + 'Segment Minimum = ' + str(min(self.value_series)) + '\n'
        description = description + 'Segment Maximum = ' + str(max(self.value_series)) + '\n'
        description = description + '---------------------------' + '\n'
        description = description + 'Segment Mean = ' + '{:.3f}'.format(i_mean) + '\n'
        description = description + 'Items Above Mean = ' + str(len(matches(bins, lambda t: t > 0))) + '\n'
        description = description + 'Items Below Mean = ' + str(len(matches(bins, lambda t: t < 0))) + '\n'
        description = description + 'Average Change = ' + '{:.3f}'.format(mean(self.mR_series)) + '\n'
        # longest run of increase
        # longest run of decrease
        description = description + '---------------------------' + '\n'
        description = description + 'Sigma(X) = ' + '{:.3f}'.format(sigma) + '\n'
        description = description + ' 3+ Sigma = (' + str(len(matches(bins, lambda t: t > 3))) + ' items)\n'
        description = description + ' 3  Sigma = ' + '{:.3f}'.format(self.individuals_standard_deviation(3)) + ' (' + str(len(matches(bins, lambda t: t == 3))) + ' items)\n'
        description = description + ' 2  Sigma = ' + '{:.3f}'.format(self.individuals_standard_deviation(2)) + ' (' + str(len(matches(bins, lambda t: t == 2))) + ' items)\n'
        description = description + ' 1  Sigma = ' + '{:.3f}'.format(self.individuals_standard_deviation(1)) + ' (' + str(len(matches(bins, lambda t: t == 1))) + ' items)\n'
        description = description + ' 0  Sigma = ' + '{:.3f}'.format(self.individuals_standard_deviation(0)) + ' (' + str(len(matches(self.value_series, lambda t: t == i_mean))) + ' items)\n'
        description = description + '-1  Sigma = ' + '{:.3f}'.format(self.individuals_standard_deviation(-1)) + ' (' + str(len(matches(bins, lambda t: t == -1))) + ' items)\n'
        description = description + '-2  Sigma = ' + '{:.3f}'.format(self.individuals_standard_deviation(-2)) + ' (' + str(len(matches(bins, lambda t: t == -2))) + ' items)\n'
        description = description + '-3  Sigma = ' + '{:.3f}'.format(self.individuals_standard_deviation(-3)) + ' (' + str(len(matches(bins, lambda t: t == -3))) + ' items)\n'
        description = description + '-3+ Sigma = (' + str(len(matches(bins, lambda t: t < -3))) + ' items)\n'
        print (description)

    def individuals_sigma_line(self, number = 1):
        value = self.individuals_standard_deviation(number)
        return [value] * len(self.value_series)

def binner(number):
    result = number // 1
    if result >= 0:
        result = result + 1
    return result
